The uid & gid are now stored using an id_pair structure.
[rsync/rsync.git] / uidlist.c
index 4daa566..b9ab1b6 100644 (file)
--- a/uidlist.c
+++ b/uidlist.c
 #include "rsync.h"
 
 #ifdef HAVE_GETGROUPS
-# if !defined(GETGROUPS_T)
+# ifndef GETGROUPS_T
 #  define GETGROUPS_T gid_t
 # endif
-# ifndef NGROUPS_MAX
-/* It ought to be defined, but just in case. */
-#  define NGROUPS_MAX 32
-# endif
 #endif
 
 extern int verbose;
@@ -50,6 +46,9 @@ struct idlist {
 static struct idlist *uidlist;
 static struct idlist *gidlist;
 
+static struct id_pair *pair_list;
+static int pair_cnt = 0, pair_alloc = 0;
+
 static struct idlist *add_to_list(struct idlist **root, int id, char *name,
                                  int id2)
 {
@@ -85,7 +84,7 @@ static char *gid_to_name(gid_t gid)
 static int map_uid(int id, char *name)
 {
        uid_t uid;
-       if (uid != 0 && name_to_uid(name, &uid))
+       if (id != 0 && name_to_uid(name, &uid))
                return uid;
        return id;
 }
@@ -93,7 +92,7 @@ static int map_uid(int id, char *name)
 static int map_gid(int id, char *name)
 {
        gid_t gid;
-       if (gid != 0 && name_to_gid(name, &gid))
+       if (id != 0 && name_to_gid(name, &gid))
                return gid;
        return id;
 }
@@ -110,9 +109,11 @@ static int is_in_group(gid_t gid)
                return last_out;
        if (ngroups < -1) {
                gid_t mygid = MY_GID();
-               if ((ngroups = getgroups(0, 0)) < 0)
+               if ((ngroups = getgroups(0, NULL)) < 0)
                        ngroups = 0;
                gidset = new_array(GETGROUPS_T, ngroups+1);
+               if (!gidset)
+                       out_of_memory("is_in_group");
                if (ngroups > 0)
                        ngroups = getgroups(ngroups, gidset);
                /* The default gid might not be in the list on some systems. */
@@ -123,8 +124,10 @@ static int is_in_group(gid_t gid)
                if (n == ngroups)
                        gidset[ngroups++] = mygid;
                if (verbose > 3) {
-                       char gidbuf[NGROUPS_MAX*16+32];
                        int pos;
+                       char *gidbuf = new_array(char, ngroups*21+32);
+                       if (!gidbuf)
+                               out_of_memory("is_in_group");
                        sprintf(gidbuf, "process has %d gid%s: ",
                            ngroups, ngroups == 1? "" : "s");
                        pos = strlen(gidbuf);
@@ -133,6 +136,7 @@ static int is_in_group(gid_t gid)
                                pos += strlen(gidbuf+pos);
                        }
                        rprintf(FINFO, "%s\n", gidbuf);
+                       free(gidbuf);
                }
        }
 
@@ -160,7 +164,7 @@ static struct idlist *recv_add_uid(int id, char *name)
        int id2 = name ? map_uid(id, name) : id;
        struct idlist *node;
 
-       node = add_to_list(&uidlist, id, name, map_uid(id, name));
+       node = add_to_list(&uidlist, id, name, id2);
 
        if (verbose > 3) {
                rprintf(FINFO, "uid %d(%s) maps to %d\n",
@@ -215,8 +219,8 @@ static gid_t match_gid(gid_t gid)
        static gid_t last_in = GID_NONE, last_out = GID_NONE;
        struct idlist *list;
 
-       if (gid == 0)
-               return 0;
+       if (gid == GID_NONE)
+               return GID_NONE;
 
        if (gid == last_in)
                return last_out;
@@ -306,7 +310,7 @@ void send_uid_list(int f)
 
 /* recv a complete uid/gid mapping from the peer and map the uid/gid
  * in the file list to local names */
-void recv_uid_list(int f, struct file_list *flist)
+void recv_uid_list(int f)
 {
        int id, i;
        char *name;
@@ -323,7 +327,6 @@ void recv_uid_list(int f, struct file_list *flist)
                }
        }
 
-
        if (preserve_gid && !numeric_ids) {
                /* read the gid list */
                while ((id = read_int(f)) != 0) {
@@ -336,14 +339,32 @@ void recv_uid_list(int f, struct file_list *flist)
                }
        }
 
/* now convert the uid/gid of all files in the list to the mapped
* uid/gid */
        if (am_root && preserve_uid && !numeric_ids) {
-               for (i = 0; i < flist->count; i++)
-                       flist->files[i]->uid = match_uid(flist->files[i]->uid);
+               for (i = 0; i < pair_cnt; i++)
+                       pair_list[i].uid = match_uid(pair_list[i].uid);
        }
        if (preserve_gid && (!am_root || !numeric_ids)) {
-               for (i = 0; i < flist->count; i++)
-                       flist->files[i]->gid = match_gid(flist->files[i]->gid);
+               for (i = 0; i < pair_cnt; i++)
+                       pair_list[i].gid = match_gid(pair_list[i].gid);
+       }
+}
+
+struct id_pair *id_pair(uid_t uid, gid_t gid)
+{
+       int i;
+
+       for (i = 0; i < pair_cnt; i++) {
+               if (uid == pair_list[i].uid && gid == pair_list[i].gid)
+                       return pair_list + i;
+       }
+
+       if (pair_cnt == pair_alloc) {
+               pair_alloc += 128;
+               pair_list = realloc_array(pair_list, struct id_pair,
+                                         pair_alloc);
        }
+       pair_list[pair_cnt].uid = uid;
+       pair_list[pair_cnt].gid = gid;
+       return pair_list + pair_cnt++;
 }
@@ -336,14 +339,32 @@ void recv_uid_list(int f, struct file_list *flist)
        }
 
-       /* now convert the uid/gid of all files in the list to the mapped
-        * uid/gid */
+       /* Now convert the id_pair array over to mapped uid/gid values. */
        if (am_root && preserve_uid && !numeric_ids) {